Object subclass: #SOLHashFrame
	instanceVariableNames: 'prev curr next'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-SOLHashTables'!
!SOLHashFrame commentStamp: 'tbr 4/4/2010 17:17' prior: 0!
I contain pointers to a contiguous segment of a linked list, to be used as a work area for traversing the list, and for performing insertion and deletion operations.  I am probably overkill for single-threaded code :), but have been left in place as a remnant of the lock-free multi-threaded list algorithm of Michaels as it is used in the SOL Hash Table of Shalev and Shavit.

In a multi-thread environment, a frame would be thread local storage, and so would probably be allocated outside of the SOLHashTable, and passed in to the various SOLHashTable operations.  That kind of interface would be very different from the more usual Squeak collections interface, so this implementation just allocates one when needed.

Tom Rushworth <tbr@acm.org>

LICENSE
--------

Copyright (c) The individual, corporate, and institutional contributors who have collectively contributed elements to this software ("The Squeak Community"), 1996-2010 All rights reserved.

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
!


!SOLHashFrame methodsFor: 'operations' stamp: 'tbr 4/1/2010 09:24'!
delete
	"Delete the node at curr from the list, and return the extracted node."

	prev next: next.
	^curr! !

!SOLHashFrame methodsFor: 'operations' stamp: 'tbr 4/4/2010 14:09'!
from: aNode find: srchKey value: srchVal
	"search down linked list of SOLHashNodes starting with (aNode next) for a node that matches srchKey and srchVal.
	 Return true if found and set  prev,curr,next to point to the 3 nodes in sequence in the list with curr being the found node.
	 Return false if not found and set prev,curr to point to the 2 nodes in sequence in the list between which the <srchKey,srchVal>
	 would have been if they were present (i.e. the insertion point if the search values are to be inserted).
	
	 This method needs to be fairly efficient, as it is called for *every* operation on the hash table."

	| currKey |
	prev := aNode.
	curr := prev next.
	[nil = curr] whileFalse:
	    ["examine the current node"
		next := curr next.
		currKey := curr key.
		(currKey = srchKey) ifTrue:
		    [ | cr |
			cr := curr compareValue: srchVal.
			(0 = cr) ifTrue: [^true].
			(0 < cr) ifTrue: [^false]
			"otherwise keep searching"]
		ifFalse:
		    [" unequal keys, quit searching if we're past the insertion point"
			currKey > srchKey ifTrue: [^false]].
		"not found yet, move the frame down the list."
		prev := curr.
		curr := next].
	^false! !

!SOLHashFrame methodsFor: 'operations' stamp: 'tbr 4/3/2010 18:45'!
insert: newNode
	"Insert newNode between prev and curr and return the inserted node"

	"Make a sanity check to see we don't ever insert a double dummy."
	"
	(newNode dummy) ifTrue:
	    [((prev key) = (newNode key)) ifTrue: [self halt].
		((nil ~= curr) and: [(curr key) = (newNode key)]) ifTrue: [self halt]].
	"
	newNode next: curr.
	prev next: newNode.
	^newNode! !


!SOLHashFrame methodsFor: 'accessing' stamp: 'tbr 4/1/2010 09:14'!
curr
	^curr! !


Association subclass: #SOLHashNode
	instanceVariableNames: 'next'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-SOLHashTables'!
!SOLHashNode commentStamp: 'tbr 4/4/2010 14:50' prior: 0!
I represent a list node in a Split-Order list.  Each item in the list has a corresponding key, and the list is kept in key order.  The key values are actually bit-reversed, to make it possible to divide the list into logical buckets, and to divide the list buckets into smaller buckets without having to re-sort the list.  It is this property that allows the SOLHashTable to expand gracefully, without ever having to rehash or rebucket the values.

There are probably many other "list" objects that would serve here, but this class is present to keep the SOLHashTable self-contained, and to make it easy to add all the bells and whistles needed for safe multi-threaded operation at some point in the future when the Squeak VM provides true independent process threads.

Tom Rushworth <tbr@acm.org>

LICENSE
---------

Copyright (c) The individual, corporate, and institutional contributors who have collectively contributed elements to this software ("The Squeak Community"), 1996-2010 All rights reserved.

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
!


!SOLHashNode methodsFor: 'accessing' stamp: 'tbr 4/2/2010 16:21'!
dummy
	"Return true if receiver is a dummy node."

	^0 = ((self key) bitAnd: 1)! !

!SOLHashNode methodsFor: 'accessing' stamp: 'tbr 4/1/2010 08:23'!
next
	^next! !

!SOLHashNode methodsFor: 'accessing' stamp: 'tbr 4/1/2010 08:23'!
next: aNode
	next := aNode! !


!SOLHashNode methodsFor: 'private' stamp: 'tbr 4/3/2010 17:00'!
compareValue: otherNode
	"Compare my value with otherNodes's value in order to disambiguate nodes with equal hash keys.
	 Return a signed integer value that corresponds roughly to the idea that (a compareValue: b)  is like (a - b) for some very loose definition of 'like'.
	 Return -1 for self < otherNode,
	           0 for self = otherNode (where '=' means 'is the same as', not identically equal),,
	          +1 for self > otherNode.
	 SortedCollection handles this with a sortBlock, that's more machinery than I want to set up for now, so just punt to a comparison method in the object that matches the one for String and implement the sortBlock later."

	| otherValue |
	otherValue := otherNode value.
	value ifNil:
	    [otherValue ifNil: [^0].
		^ -1].
	otherValue ifNil: [^1].
	^(value compare: otherValue) - 2.! !

!SOLHashNode methodsFor: 'private' stamp: 'tbr 4/1/2010 13:52'!
initialize: nextNode
	"See the class methods for proper creation and initialization."
	next := nextNode! !


!SOLHashNode methodsFor: 'search-modify' stamp: 'tbr 4/4/2010 14:16'!
delete: srchKey value: srchVal
	| frame |
	frame := SOLHashFrame new.
	(frame from: self find: srchKey value: srchVal) ifTrue:
	    ["Got it, so remove it and return the removed node."
		^frame delete].
	"Not there to delete."
	^nil! !

!SOLHashNode methodsFor: 'search-modify' stamp: 'tbr 4/4/2010 14:17'!
find: srchKey value: srchVal
	"Return the node matching srchKey and SrchVal if present."
	| frame |
	frame := SOLHashFrame new.
	(frame from: self find: srchKey value: srchVal) ifTrue: [^frame curr].
	^nil! !

!SOLHashNode methodsFor: 'search-modify' stamp: 'tbr 4/4/2010 14:17'!
insert: newNode
	"Insert newNode if not present in the list and return the newly inserted node.  Return the old node if there is already a value matching newNode in the list."

	| frame |
	frame := SOLHashFrame new.
	(frame from: self find: (newNode key) value: (newNode value)) ifTrue:
	    ["newNode's value is already present, don't insert it again, instead just return the old value."
		^frame curr]
	ifFalse:
	    ["Not found, we can insert and return the new one."
		^frame insert: newNode]! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

SOLHashNode class
	instanceVariableNames: ''!

!SOLHashNode class methodsFor: 'instance creation' stamp: 'tbr 4/1/2010 13:49'!
key: anInteger
	^(super key: anInteger value: nil) initialize: nil! !

!SOLHashNode class methodsFor: 'instance creation' stamp: 'tbr 3/25/2010 14:38'!
key: anInteger value: anObject
	^(super key: anInteger value: anObject) initialize: nil! !


Object subclass: #SOLHashSegment
	instanceVariableNames: 'depth owner pointers'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-SOLHashTables'!
!SOLHashSegment commentStamp: 'tbr 4/4/2010 14:50' prior: 0!
I represent a "segment" of a hash table containing an array of HashSegmentNumPtrs pointers to SOLHashListNodes or other segments.  The entire hash table is represented as a tree of segments of uniform depth, where the bottommost or "leaf" layer points into the list of nodes, and all higher layers point at lower segments.

depth - an int indicating position in the tree, depth == 0 implies this is a leaf segment and the pointers are to SOLHashListNodes.  This is never changed once set.

owner - the containing SOLHashTable.  Never changed once set.

pointers - an array of (owner segPtrs) pointers, which may be nil, or when depth == 0 may point to SOLHashListNodes, and which otherwise point to the next lower depth SOLHashTableSegments.

Tom Rushworth <tbr@acm.org>

LICENSE
--------

Copyright (c) The individual, corporate, and institutional contributors who have collectively contributed elements to this software ("The Squeak Community"), 1996-2010 All rights reserved.

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
!


!SOLHashSegment methodsFor: 'initialize-release' stamp: 'tbr 4/2/2010 12:31'!
initializeAtDepth: newDepth owner: anSOLHashTable
	"Set up a segment with an empty array of pointers in a tree of segments.
	 Depth is measured from the leaves of the tree, starting at 0.
	 Pointers in segments at depth 0 point to SOLHashNodes, at other depths the pointers point to segments of depth-1"

	depth := newDepth.
	owner := anSOLHashTable.
	pointers := Array new: (owner segPtrs).! !


!SOLHashSegment methodsFor: 'accessing' stamp: 'tbr 4/2/2010 12:32'!
bucketAt: bucketNum
	"Return whatever (SOLHashSegment, SOLHashNode, or nil) is in pointers at the index corresponding to bucketNum."
	^pointers at: 1+(bucketNum bitAnd: (owner segMask))! !

!SOLHashSegment methodsFor: 'accessing' stamp: 'tbr 4/2/2010 12:32'!
bucketAt: bucketNum put: anObject
	"Set the pointer array at the index corresponding to bucktNum to anObject and return anObject."
	pointers at: 1+(bucketNum bitAnd: (owner segMask)) put: anObject.
	^anObject! !

!SOLHashSegment methodsFor: 'accessing' stamp: 'tbr 4/1/2010 10:29'!
depth
	^depth! !


Collection subclass: #SOLHashTable
	instanceVariableNames: 'approxLoadFactor numElements keyMask segBits segPtrs segMask buckets hashBlock keyBits'
	classVariableNames: 'Elbbyn'
	poolDictionaries: ''
	category: 'Collections-SOLHashTables'!
!SOLHashTable commentStamp: 'tbr 4/4/2010 17:13' prior: 0!
This is released under the Squeak license, see the LICENSE section below.

Background:
--------------

An SOLHashTable is an expandable hash table based on Split-Ordered Lists, as described in the paper
   "Split-Ordered Lists: Lock-Free Extensible Hash Tables"
by
    Ori Shalev and Nir Shavit,
in
   Journal of the ACM, Vol. 53, No. 3, May 2006, pp 379-405.

The algorithm as presented in the paper is a multi-thread-safe lock-free version that uses CAS, and is one of the prettiest algorithms I've seen in a long time. Since Squeak is not at present multi-threaded, the algorithm can be simplified by using a plain store instead of CAS, removing the re-try logic on CAS failure, and removing all of the thread-safe list deletion code. This class and its compaions represents the simplified version.

Note: objects added to this implementation must respond to 'compare: anObject' in the same way Strings do.  See the Collisions section below for more detail.  Feel free to find a better way to deal with key collisions.

The actual keys and values are kept in a linked list of SOLHashNodes, and the list is managed by keeping a tree of SOLHashSegments whose leaves are pointers at dummy nodes in the list.  Each portion of the list between dummy nodes corresponds to a bucket in a traditional hash table.  There is no explicit list head in the SOLHashTable, the bottommost or lowest leaf in the lowest table segment points at the dummy list node that operates as the list head.

Collisions:
-----------

There are two types of things generally called hash collisions, (1) when two different keys are assigned to the same bucket (bucket collision), and (2) when two different objects get the same hash key (key collision).  This algorithm expects bucket collisions and handles them with the split-ordered-list. Key collisions are best handled in different ways for different scenarios.  Some types of collection simply rule them out, others allow them but have different ways of resolving ambiguities.  This implementation does not try to provide the best fitted solution, it just uses an auxilliary object supplied in both the find and delete methods to resolve the ambiguities.  This auxilliary object does NOT need to be the same object as the one stored in the table, it just needs to 'compare:' as equal.

Weaknesses
-------------

This data structure MUST have a good hash function, where all bits of the hash keys have approximately the same likelyhood of changing between different hashed objects.  You don't get to sprinkle magic primality dust over it after the hash and have it work well :).  I tried the 12 bit Squeak native hash on a very small file of 30 or so words, and ended up with 11 of the objects in the same bucket.  Googling for string string hashing turns up all sorts of candidate functions, the second one I tried worked pretty well for my test domain of English text.  If you want to use this algorithm, you will need to try different hash functions out on your data.  Use the showBucketSizeHistgram method to see how well a particular function works for your data.

Miscellaneous bits of stuff used during development:
--------------------------------------------------------------
| tbl |
tbl := SOLHashTable new loadFactor: 4 segBits: 4.
(SOLHashTable hash32: 'the') printStringRadix: 16 '16rDBC3966D'
(SOLHashTable bitReverse32: 16rDBC3966D) printStringRadix: 16 '16rB669C2DB'
tbl at: (tbl hash: 'the') put: 'the'
tbl find: (tbl hash: 'the') value: 'the'
tbl showNodeList
tbl showBucketSizeHistogram
tbl delete: (tbl hash: 'the') value: 'the'

(SOLHashTable new loadFactor: 4 segBits: 4) explore
self at: (self hash: 'the') put: 'the'
self find: (self hash: 'the') value: 'the' 'the'
self showNodeList
self showBucketSizeHistogram
self delete: (self hash: 'the') value: 'the'

(SOLHashTable loadFromFile: '/Users/tbr/Documents/Books/Originals/Austen/nabby11.txt') explore
(SOLHashTable loadFromFile: '/Users/tbr/IdleWork/SOL_Hash/common.h') explore

Original port to Squeak by: Tom Rushworth <tbr@acm.org>

LICENSE
---------

Copyright (c) The individual, corporate, and institutional contributors who have collectively contributed elements to this software ("The Squeak Community"), 1996-2010 All rights reserved.

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
!
]style[(3225 109 12 89 4 189 5 23 4 398 1213)f2,f1,f2,f1,f2,f1,f2,f1,f2,f1,f2!


!SOLHashTable methodsFor: 'initialize-release' stamp: 'tbr 4/3/2010 18:35'!
loadFactor: loadFactor segBits: numBits
	"This initializer supplies a default hashBlock that is expecting a ByteString.
	 If you use it, and are not adding ByteStrings, don't use the 'put:', 'find:' or 'delete:' methods, use the
	 methods where you supply the key yourself."
	self loadFactor: loadFactor segBits: numBits hash: [ :str | self hash: str ].! !

!SOLHashTable methodsFor: 'initialize-release' stamp: 'tbr 4/4/2010 08:48'!
loadFactor: loadFactor segBits: numBits hash: aBlock
	approxLoadFactor := loadFactor.
	segBits := numBits.
	segPtrs := 1 bitShift: numBits.
	segMask := segPtrs - 1.
	keyMask := 1.
	numElements := 0.
	"Uses the Squeak native hash, but the native hash has poor bit distribution."
	"keyBits := 12."
	"Uses the low-order 28 bits of the 32-bit Murmur2 hash, plus 1 bit for dummy/regular for 29 bits, still SmallInteger friendly.
	 This works reasonably well for ByteStrings of English text."
	keyBits := 28.
	"Uses the low-order 31 bits of the 32-bit Murmur2 hash, plus 1 bit for dummy/regular for a total of 32 bits.
	 Thiw works reasonably well, but is slower."
	"keyBits := 31."
	hashBlock := aBlock.
	self initializeBucket: 0.! !


!SOLHashTable methodsFor: 'searching-modifying' stamp: 'tbr 4/3/2010 18:32'!
at: key put: value
	"Add a value for key in the hash table.  Return nil if successful, or the old value if key is already present."

	| bucketNum newNode hn oldNode |
	bucketNum := key bitAnd: keyMask.
	newNode := SOLHashNode key: (self regularKey: key) value: value.
	hn := self bucketAt: bucketNum.
	(nil = hn) ifTrue:
	    ["Bucket bucketNum has not yet been initialized, so do it now."
		hn := self initializeBucket: bucketNum].
	oldNode := hn insert: newNode.
	(oldNode == newNode) ifFalse:
	    ["key and value were already in the table, don't add anything new, just return the old value."
		^oldNode value].
	"newNode successfully added."
	numElements := numElements + 1.
	"Now consider expanding the table, based on numElements, approxLoadFactor and number of buckets (from keyMask)."
	(numElements > ((keyMask + 1) * approxLoadFactor)) ifTrue:
	    ["Expanding the table looks like a good idea."
		self widen.].
	^value! !

!SOLHashTable methodsFor: 'searching-modifying' stamp: 'tbr 4/3/2010 17:32'!
delete: srchVal
	"Remove the value from the table and return it if present."

	^self delete: (hashBlock value: srchVal) value: srchVal! !

!SOLHashTable methodsFor: 'searching-modifying' stamp: 'tbr 4/3/2010 18:33'!
delete: srchKey value: srchVal
	"Remove the value matching srchKey and srchVal and return it if present."

	| bucketNum hn oldNode |
	bucketNum := srchKey bitAnd: keyMask.
	hn := self bucketAt: bucketNum.
	(nil = hn) ifTrue: [^nil].
	oldNode := hn delete: (self regularKey: srchKey) value: srchVal.
	(nil = oldNode) ifTrue:[^nil].
	numElements := numElements - 1.
	^oldNode value! !

!SOLHashTable methodsFor: 'searching-modifying' stamp: 'tbr 4/3/2010 17:33'!
find: srchVal 
	"Return the stored value matching srchVal if present, otherwise nil."

	^self find: (hashBlock value: srchVal) value: srchVal! !

!SOLHashTable methodsFor: 'searching-modifying' stamp: 'tbr 4/3/2010 18:33'!
find: srchKey value: srchVal 
	"Return the stored value matching srchKey and srchVal if present, otherwise nil."

	| bucketNum hn oldNode |
	bucketNum := srchKey bitAnd: keyMask.
	hn := self bucketAt: bucketNum.
	(nil = hn) ifTrue: [^nil].
	oldNode := hn find: (self regularKey: srchKey) value: srchVal.
	(nil = oldNode) ifTrue: [^nil].
	^oldNode value! !

!SOLHashTable methodsFor: 'searching-modifying' stamp: 'tbr 4/3/2010 17:29'!
put: value
	"Add a value to the hash table with the hash key obtained from value by the hashBlock.
	 Return nil if successful, or the old value if the key is already present."

	^self at: (hashBlock value: value) put: value! !


!SOLHashTable methodsFor: 'accessing' stamp: 'tbr 4/3/2010 11:49'!
height
	buckets ifNil: [^0].
	^(buckets depth) + 1! !

!SOLHashTable methodsFor: 'accessing' stamp: 'tbr 4/3/2010 11:48'!
keyMask
	^keyMask! !

!SOLHashTable methodsFor: 'accessing' stamp: 'tbr 4/2/2010 12:29'!
segBits
	^segBits! !

!SOLHashTable methodsFor: 'accessing' stamp: 'tbr 4/2/2010 12:30'!
segMask
	^segMask! !

!SOLHashTable methodsFor: 'accessing' stamp: 'tbr 4/2/2010 12:29'!
segPtrs
	^segPtrs! !

!SOLHashTable methodsFor: 'accessing' stamp: 'tbr 4/1/2010 13:10'!
size
	^numElements! !


!SOLHashTable methodsFor: 'enumerating' stamp: 'tbr 4/2/2010 13:52'!
allNodesDo: aBlock
	"Evaluate aBlock once for each node, both regular and dummy."

	| node |
	node := self bucketAt: 0.
	[nil = node] whileFalse:
	    [aBlock value: node.
		node := node next]! !

!SOLHashTable methodsFor: 'enumerating' stamp: 'tbr 4/2/2010 14:47'!
associationsDo: aBlock
	"Evaluate aBlock once for each regular node value."

	self regularNodesDo: aBlock! !

!SOLHashTable methodsFor: 'enumerating' stamp: 'tbr 4/2/2010 14:45'!
do: aBlock
	"Evaluate aBlock for each of the receiver's regular values."

	self valuesDo: aBlock! !

!SOLHashTable methodsFor: 'enumerating' stamp: 'tbr 4/2/2010 14:48'!
dummyNodesDo: aBlock
	"Evaluate aBlock once for each dummy node value."

	self allNodesDo: [ :node | (node dummy) ifTrue: [aBlock value: node]]! !

!SOLHashTable methodsFor: 'enumerating' stamp: 'tbr 4/3/2010 18:29'!
keysAndValuesDo: aBlock
	"Evaluate aBlock once for each of the receiver's <regular key, value> pairs, converting the keys back to the original hash keys."

	self regularNodesDo: [ :node | aBlock value: (self hashKey: node key) value: node value]! !

!SOLHashTable methodsFor: 'enumerating' stamp: 'tbr 4/3/2010 18:30'!
keysDo: aBlock
	"Evaluate aBlock once for each of the receiver's regular keys, converting the keys back to the original hash keys."

	self regularNodesDo: [ :node | aBlock value: (self hashKey: node key)]! !

!SOLHashTable methodsFor: 'enumerating' stamp: 'tbr 4/2/2010 14:49'!
regularNodesDo: aBlock
	"Evaluate aBlock once for each regular node value."

	self allNodesDo: [ :node | (node dummy) ifFalse: [aBlock value: node]]! !

!SOLHashTable methodsFor: 'enumerating' stamp: 'tbr 4/2/2010 14:44'!
valuesDo: aBlock
	"Evaluate aBlock once for each of the receiver's values."

	self regularNodesDo: [ :node | aBlock value: node value]! !


!SOLHashTable methodsFor: 'testing' stamp: 'tbr 4/3/2010 18:14'!
pathForDummy: dummyNode
	"Return a string containg a comma separated list of SOLHashSegment indexes leading to the bucket starting at dummyNode.
	 If you try this on a regular node you may hit uninitialized buckets or segments."

	| frwdKey seg depth path sx |
	frwdKey := self bitReverse: (dummyNode key).
	seg := buckets.
	(nil = seg) ifTrue:
	    [depth := 0.
		path := '']
	ifFalse:
	    [depth := seg depth.
		sx := (frwdKey bitShift: ((segBits * depth) negated)) bitAnd: segMask.
		path := sx printString].
	[0 < depth] whileTrue:
	    [depth := depth - 1.
		seg := seg bucketAt: sx.
		sx := (frwdKey bitShift: ((segBits * depth) negated)) bitAnd: segMask.
		path := path, ',', (sx printString)].
	^path! !

!SOLHashTable methodsFor: 'testing' stamp: 'tbr 4/2/2010 16:43'!
showBucketSizeHistogram
	"Write a bucket size histogram to the transcript."

	| histogram numDummies numNodes |
	histogram := Dictionary new.
	numDummies := 0.
	numNodes := 0.
	"Run through the node list, counting bucket sizes."
	self allNodesDo:
	    [ :node |
	     (node dummy) ifTrue:
		    [(numDummies > 0) ifTrue:
			    ["We're at the end of a bucket, add an observation to the histogram."
				histogram at: numNodes put: ((histogram at: numNodes ifAbsent: [0]) + 1)].
			numNodes := 0.
			numDummies := numDummies + 1]
		ifFalse:
		    [numNodes := numNodes + 1]].
	(numDummies > 0) ifTrue:
	    ["We're at the end of the last bucket, add an observation to the histogram."
		histogram at: numNodes put: ((histogram at: numNodes ifAbsent: [0]) + 1)].
	"We have the histogram data, now display it."
	Transcript cr; show: '---------- SOLHashTable bucket size histogram, ';
	             show: numDummies printString; show: ' buckets, ';
	             show: numElements printString; show: ' values'; cr;
	             show: ' size :  count'; cr..
	(histogram keysSortedSafely) do:
	    [ :key |
		Transcript show: (key printPaddedWith: $  to: 5); show: ' :'; show: ((histogram at: key) printPaddedWith: $  to: 8); cr].! !

!SOLHashTable methodsFor: 'testing' stamp: 'tbr 4/3/2010 18:30'!
showNodeList
	"Write the entire list of nodes to the transcript.
	 It's not a good idea to do this for large tables."

	| seg depth nx |
	seg := buckets.
	(nil = seg) ifTrue: [depth := 0] ifFalse: [depth := 1 + seg depth].
	Transcript cr; show: '---------- SOLHashTable height='; show: depth; show: ' size='; show: self size; cr.
	nx := 0.
	self allNodesDo:
	    [ :node |
		(node dummy) ifTrue:
		    ["Show the dummy node, no value."
			Transcript show: nx; show: ' hash='; show: ((self hashKey: (node key)) printPaddedWith: $0 to: 8 base: 16);
			             show: ' rvrs='; show: (node key printPaddedWith: $0 to: 8 base: 16); show: ' dummy, [', (self pathForDummy: node), ']';
			             cr]
		ifFalse:
		    ["Show a regular node."
			Transcript show: nx; show: ' hash='; show: ((self hashKey: (node key)) printPaddedWith: $0 to: 8 base: 16);
			             show: ' rvrs='; show: (node key printPaddedWith: $0 to: 8 base: 16); show: ' -> ';
			             show: node value printString;
			             cr].
		nx := nx + 1]! !


!SOLHashTable methodsFor: 'private' stamp: 'tbr 4/4/2010 08:32'!
bitReverse: anUnsignedInt
	(28 = keyBits) ifTrue: [^self class bitReverse29: anUnsignedInt].
	(12 = keyBits) ifTrue: [^self class bitReverse13: anUnsignedInt].
	^self class bitReverse32: anUnsignedInt! !

!SOLHashTable methodsFor: 'private' stamp: 'tbr 4/3/2010 17:05'!
bucketAt: bucketNum
	"Return the dummy SOLHashNode that is the head of bucket bucketNum.
	 Return nil if the containing segment is not yet present or the bucket has not been initialized."

	| seg depth |
	seg := buckets.
	[nil = seg] whileFalse:
	    [depth := seg depth.
		(0 = depth) ifTrue:
		    ["We're at the bottom level, this segment contains pointers to nodes.
			 Return the one we are interested in."
			^seg bucketAt: bucketNum].
		"Not at the bottom, this segment points at segments with depth one less."
		seg := seg bucketAt: (bucketNum bitShift: ((depth * segBits) negated))].
	"We ran out of segments before reaching the bottom."
	^nil! !

!SOLHashTable methodsFor: 'private' stamp: 'tbr 4/2/2010 12:35'!
bucketAt: bucketNum putNode: node
	"Set the dummy SOLHashNode that is the head of bucket bucketNum and return the node.
	 Instantiate any missing segments needed to do so."

	| seg depth parentSeg parentNdx |
	seg := buckets.
	parentSeg := nil.
	parentNdx := 0.
	depth := 0.
	[true] whileTrue:
	    ["step down through the segments until we reach the bottom."
		(nil = seg) ifTrue:
		    ["Missing segment, fill it in."
			seg := SOLHashSegment new initializeAtDepth: depth owner: self.
			(nil = parentSeg) ifTrue:
			    [buckets := seg]
			ifFalse:
			    [seg := parentSeg bucketAt: parentNdx put: seg]].
		depth := seg depth.
		(0 = depth) ifTrue:
		    ["We're at the bottom level, this segment contains pointers to nodes.
			 Return the one we are interested in."
			^seg bucketAt: bucketNum put: node].
		"Not at the bottom, this segment points at segments with depth one less."
		parentSeg := seg.
		parentNdx := bucketNum bitShift: ((depth * segBits) negated).
		seg := parentSeg bucketAt: parentNdx.
		depth := depth - 1].
	"We ran out of segments before reaching the bottom."
	^nil! !

!SOLHashTable methodsFor: 'private' stamp: 'tbr 4/3/2010 17:11'!
deepen
	"Deepen the segment tree by one layer, putting the old tree in bucket 0 of the new segment."
	
	| oldBuckets newBuckets |
	oldBuckets := buckets.
	newBuckets := SOLHashSegment new initializeAtDepth: (oldBuckets depth)+1 owner: self.
	newBuckets bucketAt: 0 put: oldBuckets.
	buckets := newBuckets.
	! !

!SOLHashTable methodsFor: 'private' stamp: 'tbr 4/4/2010 08:32'!
dummyKey: frwdKey
	(28 = keyBits) ifTrue: [^self class dummyKey29: frwdKey].
	(12 = keyBits) ifTrue: [^self class dummyKey13: frwdKey].
	^self class dummyKey32: frwdKey! !

!SOLHashTable methodsFor: 'private' stamp: 'tbr 4/4/2010 08:54'!
hash: aByteString
	(28 = keyBits) ifTrue: [^self class hash28: aByteString].
	(12 = keyBits) ifTrue: [^self class hash12: aByteString].
	^self class hash31: aByteString! !

!SOLHashTable methodsFor: 'private' stamp: 'tbr 4/4/2010 08:34'!
hashKey: rvrsKey
	(28 = keyBits) ifTrue: [^self class hashKey29: rvrsKey].
	(12 = keyBits) ifTrue: [^self class hashKey13: rvrsKey].
	^self class hashKey32: rvrsKey! !

!SOLHashTable methodsFor: 'private' stamp: 'tbr 4/3/2010 18:31'!
initializeBucket: bucketNum
	"Each bucket in the table points at a dummy SOLHashNode, so we want to create the dummy and insert it into the appropriate spot.
	 We look for the parent bucket to find the closest place to perform the insertion.
	 The very first bucket is a special case because there is neither parent nor segments yet."

	| parent dummy parentNum |
	parent := nil.
	parentNum := 0.
	(0 ~= bucketNum) ifTrue:
	    ["There is a parent node to look for."
		parentNum := SOLHashTable bitParent: bucketNum.
		parent := self bucketAt: parentNum.
		(nil = parent) ifTrue:
		    ["Parent bucket has not been initialized yet either,  Do so recursively."
			parent := self initializeBucket: parentNum]].
	"Create the dummy node."
	dummy := SOLHashNode key: (self dummyKey: bucketNum).
	(nil ~= parent) ifTrue:
	    ["Insert the new dummy node in the parent list."
		dummy := parent insert: dummy].
	"Transcript cr; show: '+++++ initiaizing bucket ',(bucketNum printString),' key#',(dummy key printStringRadix:16); cr."
	^self bucketAt: bucketNum putNode: dummy
	! !

!SOLHashTable methodsFor: 'private' stamp: 'tbr 4/4/2010 08:35'!
regularKey: frwdKey
	(28 = keyBits) ifTrue: [^self class regularKey29: frwdKey].
	(12 = keyBits) ifTrue: [^self class regularKey13: frwdKey].
	^self class regularKey32: frwdKey! !

!SOLHashTable methodsFor: 'private' stamp: 'tbr 4/3/2010 17:11'!
widen
	"Double the number of available buckets.
	 This does not initialize any of the new buckets, that happens lazily, as they are filled."

	| oldBuckets depth maxBuckets numBuckets |
	"Before we can actually double the number of buckets, we need to see if the doubled number will fit into our current segment tree."
	oldBuckets := buckets.
	depth := oldBuckets depth.
	maxBuckets := 1 bitShift: segBits*(depth+1).
	numBuckets := 2*(keyMask+1).
	(maxBuckets < numBuckets) ifTrue:
	    ["We need a deeper segment tree."
		self deepen].
	"The segment tree is big enough, double the available buckets by adding a bit to the keyMask."
	keyMask := numBuckets - 1.
	! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

SOLHashTable class
	instanceVariableNames: ''!

!SOLHashTable class methodsFor: 'key-construction' stamp: 'tbr 4/4/2010 08:25'!
dummyKey13: frwdKey
	"Construct a dummy list node key from a hash segment key."

	^self bitReverse13: (frwdKey bitAnd: 16r0FFF)! !

!SOLHashTable class methodsFor: 'key-construction' stamp: 'tbr 4/4/2010 08:26'!
dummyKey29: frwdKey
	"Construct a dummy list node key from a hash segment key."

	^self bitReverse29: (frwdKey bitAnd: 16rFFFFFFF)! !

!SOLHashTable class methodsFor: 'key-construction' stamp: 'tbr 4/4/2010 08:25'!
dummyKey32: frwdKey
	"Construct a dummy list node key from a hash segment key."

	^self bitReverse32: (frwdKey bitAnd: 16r7FFFFFFFF)! !

!SOLHashTable class methodsFor: 'key-construction' stamp: 'tbr 4/4/2010 08:27'!
hashKey13: rvrsKey
	"Construct a hash key from a list node key (where list node keys are bitReversed).
	 This method strips off the 'regular key' bit, so you can't distinguish regular keys from dummies."

	^((self bitReverse13: rvrsKey) bitAnd: 16rFFF) as31BitSmallInt! !

!SOLHashTable class methodsFor: 'key-construction' stamp: 'tbr 4/3/2010 18:22'!
hashKey29: rvrsKey
	"Construct a hash key from a list node key (where list node keys are bitReversed).
	 This method strips off the 'regular key' bit, so you can't distinguish regular keys from dummies."

	^((self bitReverse29: rvrsKey) bitAnd: 16rFFFFFFF) as31BitSmallInt! !

!SOLHashTable class methodsFor: 'key-construction' stamp: 'tbr 4/3/2010 18:23'!
hashKey32: rvrsKey
	"Construct a hash key from a list node key (where list node keys are bitReversed).
	 This method strips off the 'regular key' bit, so you can't distinguish regular keys from dummies."

	^(self bitReverse32: rvrsKey) bitAnd: 16r7FFFFFFF! !

!SOLHashTable class methodsFor: 'key-construction' stamp: 'tbr 4/4/2010 08:29'!
regularKey13: frwdKey
	"Construct a regular list node key from a hash segment key."

	^self bitReverse13: (frwdKey bitOr: 16r1000)! !

!SOLHashTable class methodsFor: 'key-construction' stamp: 'tbr 4/4/2010 08:15'!
regularKey29: frwdKey
	"Construct a regular list node key from a hash segment key."

	^self bitReverse29: ((frwdKey bitOr: 16r10000000) as31BitSmallInt)! !

!SOLHashTable class methodsFor: 'key-construction' stamp: 'tbr 4/3/2010 18:15'!
regularKey32: frwdKey
	"Construct a regular list node key from a hash segment key."

	^(self bitReverse32: frwdKey) bitOr: 1! !


!SOLHashTable class methodsFor: 'bit-manipulation' stamp: 'tbr 4/1/2010 18:08'!
bitParent: anUnsignedInteger
	"Return anUnsignedInteger with the most significant 1-bit changed to 0."
	(0 = anUnsignedInteger) ifTrue: [^0].
	^ anUnsignedInteger bitXor: (1 bitShift: ((anUnsignedInteger highBit) - 1))! !

!SOLHashTable class methodsFor: 'bit-manipulation' stamp: 'tbr 4/4/2010 08:19'!
bitReverse13: anUnsignedInteger
	"Return the low order 13-bit reversed value of anUnsignedInteger.
	 Someone with more Squeak skill than I should improve this.
	 The C version of this uses a table of 256 bit reversed bytes, I'm using 4-bit nybbles here because I don't want to enter the whole table."

	| rvrs n frwd |
	(nil = Elbbyn) ifTrue:
	    [Elbbyn := #(0 8 4 12 2 10 6 14 1 9 5 13 3 11 7 15)].
	frwd := anUnsignedInteger.
	rvrs := 0.
	0 to: 2 do:
	    [:i | n := frwd bitAnd: 16rF.
		frwd := frwd bitShift: -4.
		n := Elbbyn at: (1+n).
		rvrs := (rvrs bitShift: 4) bitOr: n].
	"12 bits done, now the 13th."
	rvrs := (rvrs bitShift: 1) bitOr: (frwd bitAnd: 1).
	^rvrs as31BitSmallInt! !

!SOLHashTable class methodsFor: 'bit-manipulation' stamp: 'tbr 4/3/2010 18:10'!
bitReverse29: anUnsignedInteger
	"Return the low order 29-bit reversed value of anUnsignedInteger.
	 Someone with more Squeak skill than I should improve this.
	 The C version of this uses a table of 256 bit reversed bytes, I'm using 4-bit nybbles here because I don't want to enter the whole table."

	| rvrs n frwd |
	(nil = Elbbyn) ifTrue:
	    [Elbbyn := #(0 8 4 12 2 10 6 14 1 9 5 13 3 11 7 15)].
	frwd := anUnsignedInteger.
	rvrs := 0.
	0 to: 6 do:
	    [:i | n := frwd bitAnd: 16rF.
		frwd := frwd bitShift: -4.
		n := Elbbyn at: (1+n).
		rvrs := (rvrs bitShift: 4) bitOr: n].
	"28 bits done, now the 29th."
	rvrs := (rvrs bitShift: 1) bitOr: (frwd bitAnd: 1).
	^rvrs as31BitSmallInt! !

!SOLHashTable class methodsFor: 'bit-manipulation' stamp: 'tbr 4/3/2010 18:10'!
bitReverse32: anUnsignedInteger
	"Return the 32-bit reversed value of anUnsignedInteger.
	 Someone with more Squeak skill than I should improve this.
	 The C version of this uses a table of 256 bit reversed bytes, I'm using 4-bit nybbles here because I don't want to enter the whole table."

	| rvrs n frwd |
	(nil = Elbbyn) ifTrue:
	    [Elbbyn := #(0 8 4 12 2 10 6 14 1 9 5 13 3 11 7 15)].
	frwd := anUnsignedInteger.
	rvrs := 0.
	0 to: 7 do:
	    [:i | n := frwd bitAnd: 16rF.
		frwd := frwd bitShift: -4.
		n := Elbbyn at: (1+n).
		rvrs := (rvrs bitShift: 4) bitOr: n].
	^rvrs! !


!SOLHashTable class methodsFor: 'testing' stamp: 'tbr 4/4/2010 08:37'!
hash12: aByteString
	"Use the Squeak native 12-bit hash."
	^(aByteString hash) as31BitSmallInt! !

!SOLHashTable class methodsFor: 'testing' stamp: 'tbr 4/3/2010 18:04'!
hash28: aByteString
	"Calculate the 32-bit MurmurHashNeutral2 on aString, then chop it to 28 bits in order to be SmallInteger friendly."

	^((self hash32: aByteString) bitAnd: 16rFFFFFFF) as31BitSmallInt! !

!SOLHashTable class methodsFor: 'testing' stamp: 'tbr 4/4/2010 08:51'!
hash31: aByteString
	"Calculate MurmurHashNeutral2 on aString."

	^(self hash32: aByteString) bitAnd: 16r7FFFFFFF! !

!SOLHashTable class methodsFor: 'testing' stamp: 'tbr 4/3/2010 17:59'!
hash32: aByteString
	"Calculate MurmurHashNeutral2 on aString."

	| seed len len4 h k i m r |
	seed := 16rAAAAAAAA.
	m := 16r5BD1E995.
	r := -24.
	len := aByteString size.
	len4 := len bitAnd: 16rFFFC.
	h := seed bitXor: len.
	i := 0.
	[i < len4] whileTrue:
         [k := aByteString byteAt: i+1.
	     k := (k bitShift: 8) bitOr: (aByteString byteAt: i+2).
		k := (k bitShift: 8) bitOr: (aByteString byteAt: i+3).
		k := (k bitShift: 8) bitOr: (aByteString byteAt: i+4).
		
		k := (k * m) bitAnd: 16rFFFFFFFF.
		k := k bitXor: (k bitShift: r).
		k := (k * m) bitAnd: 16rFFFFFFFF.
		
		h := ((h * m) bitAnd: 16rFFFFFFFF) bitXor: k.
		i := i + 4].
	"Now deal with up to 3 bytes in the tail end."
	len := len bitAnd: 3.
	(len >= 3) ifTrue: [h := h bitXor: ((aByteString byteAt: i+3) bitShift: 16)].
	(len >= 2) ifTrue: [h := h bitXor: ((aByteString byteAt: i+2) bitShift: 8)].
	(len >= 1) ifTrue: [h := h bitXor: (aByteString byteAt: i+1).
		h := (h * m) bitAnd: 16rFFFFFFFF.].
	
	h := h bitXor: (h bitShift: -13).
	h := (h * m) bitAnd: 16rFFFFFFFF.
	h := h bitXor: (h bitShift: -15).
	^h! !

!SOLHashTable class methodsFor: 'testing' stamp: 'tbr 4/3/2010 18:05'!
loadFromFile: fileName
	"Create an instance and load it with strings from fileName."

	| tbl fileStream line whiteSpace wn checkWord currKeyMask currHeight currSize |
	checkWord := 0 "77".
	whiteSpace := ' ',(Character tab asString),(Character cr asString),(Character lf asString).
	tbl := SOLHashTable new loadFactor: 6 segBits: 4.
	currKeyMask := tbl keyMask.
	currHeight := tbl height.
	currSize := tbl size.
	fileStream := FileStream readOnlyFileNamed: fileName.
	wn := 0.
	[nil = (line := fileStream nextLine)] whileFalse:
	    [(line asByteString findTokens: whiteSpace) do:
		    [ :word |
			wn := wn + 1.
			"Transcript show: (wn printString),' word={',word,'}'; cr."
			"(checkWord = wn) ifTrue: [self halt]."
			"tbl at: (SOLHashTable hash32: word) put: word."
			tbl put: word.
			(checkWord = wn) ifTrue: [tbl showNodeList].
			"
			(currSize = (tbl size)) ifFalse:
			    [currSize := tbl size.
				((currSize >= 60) and: [currSize <= 61]) ifTrue:
				    [tbl showBucketSizeHistogram]].
			"
			"
			(currKeyMask = (tbl keyMask)) ifFalse:
			    [currKeyMask := tbl keyMask.
				(currHeight = (tbl height)) ifFalse:
				    [currHeight := tbl height.
					Transcript cr; show: '>>>>>>>>>>>>>> table deepened, height now ',(currHeight printString); cr]
				ifTrue:
				    [Transcript cr; show: '>>>>>>>>>>>>>> table widened, mask now ',(currKeyMask printStringRadix: 16); cr].
				tbl showBucketSizeHistogram]
			"
			]].
	fileStream close.
	^tbl! !
